#include "initguid.h"
#include "windows.h"
#include "commctrl.h"
#include "gdiplus.h"
#include "setupapi.h"
#include "dbt.h"
#include "udisk.h"
#include "serial.h"
#include "pixels.h"
#include "utilfunc.h"
#include "resource.h"



HINSTANCE  g_hInstance;
HWND       g_hDlg;
HWND       g_hBmp;
HDEVNOTIFY g_hDevNotify;
LPTSTR    *g_DevList;
DWORD      g_DevListLength;
HANDLE     g_hThread;
BOOL       g_bProcessing;
BOOL       g_bClose;
LPTSTR     g_lpDevice;
DWORD      g_nScreen;
TCHAR      g_lpBaseName[MAX_PATH];

BOOL  CALLBACK MainDlgProc(HWND, UINT, WPARAM, LPARAM);
BOOL  CALLBACK BmpWindowProc(HWND, UINT, WPARAM, LPARAM);
BOOL  CALLBACK AboutDlgProc(HWND, UINT, WPARAM, LPARAM);
DWORD WINAPI ThreadProc(LPVOID);
BOOL  DoRegisterDeviceInterface(GUID InterfaceClassGuid, HDEVNOTIFY *hDevNotify);
void  ListDevices();
void  ListScreens();
void  DestroyList();
void  OnInitDialog(HWND);
BOOL  OnCommand(WPARAM);
void  OnPaint(HDC);

//{2c7089aa-2e0e-11d1-b114-00c04fc2aae4}
DEFINE_GUID(GUID_DEVINTERFACE_MODEM,
	0x2c7089aa, 0x2e0e, 0x11d1, 0xb1, 0x14, 0x00, 0xc0, 0x4f, 0xc2, 0xaa, 0xe4);


int WINAPI WinMain(HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPSTR lpCmdLine,
					int nCmdShow)
{
	Log(
		TEXT("MSnap 0.2b ")
#ifdef _DEBUG
		TEXT("DEBUG ")
#endif
#ifdef UNICODE
		TEXT("UNICODE ")
#endif
		TEXT("built on ")TEXT(__DATE__)TEXT(" ")TEXT(__TIME__)
	);
	g_hInstance = GetModuleHandle(NULL);

	Gdiplus::GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR           gdiplusToken;
	Gdiplus::GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);

/*	PROPSHEETPAGE psp[2];
	PROPSHEETHEADER psh;

	psp[0].dwSize        = sizeof(PROPSHEETPAGE);
	psp[0].dwFlags       = PSP_DEFAULT;
	psp[0].hInstance     = g_hInstance;
	psp[0].pszTemplate   = MAKEINTRESOURCE(IDD_MAINDLG);
	psp[0].pszIcon       = NULL;
	psp[0].pfnDlgProc    = MainDlgProc;
	psp[0].lParam        = 0;

	psp[1].dwSize        = sizeof(PROPSHEETPAGE);
	psp[1].dwFlags       = PSP_DEFAULT;
	psp[1].hInstance     = g_hInstance;
	psp[1].pszTemplate   = MAKEINTRESOURCE(IDD_ABOUTDLG);
	psp[1].pszIcon       = NULL;
	psp[1].pfnDlgProc    = AboutDlgProc;
	psp[1].lParam        = 0;

	psh.dwSize           = sizeof(PROPSHEETHEADER);
	psh.dwFlags          = PSH_PROPSHEETPAGE | PSH_USEICONID;
	psh.hwndParent       = NULL;
	psh.hInstance        = g_hInstance;
	psh.pszIcon          = MAKEINTRESOURCE(IDI_MAINICON);
	psh.pszCaption       = TEXT("MSnap");
	psh.nPages           = sizeof(psp) / sizeof(PROPSHEETPAGE);
	psh.ppsp             = (LPCPROPSHEETPAGE) &psp;

	PropertySheet(&psh);
*/

	InitCommonControls();

	DialogBox(g_hInstance, MAKEINTRESOURCE(IDD_MAINDLG), NULL, MainDlgProc);

	Gdiplus::GdiplusShutdown(gdiplusToken);

	Log(TEXT("************************"));
	ExitProcess(0);
}


BOOL CALLBACK MainDlgProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	PDEV_BROADCAST_HDR lpdb;

	switch (uMsg)
	{
	case WM_COMMAND:
		return OnCommand(wParam);
		break;

	case WM_DEVICECHANGE:
		lpdb = (PDEV_BROADCAST_HDR)lParam;
		switch (wParam) {
		case DBT_DEVICEARRIVAL:
			switch (lpdb -> dbch_devicetype) {
			case DBT_DEVTYP_DEVICEINTERFACE:
				Log(TEXT("WM_DEVICECHANGE::DBT_DEVICEARRIVAL::DBT_DEVTYP_DEVICEINTERFACE"));
				ListDevices();
				break;
			case DBT_DEVTYP_VOLUME:
				Log(TEXT("WM_DEVICECHANGE::DBT_DEVICEARRIVAL::DBT_DEVTYP_VOLUME"));
				UDiskToMode(AT_Mode_Modem);
				break;
			default:
				Log(TEXT("WM_DEVICECHANGE::DBT_DEVICEARRIVAL"));
				return FALSE;
			}
			break;

		case DBT_DEVICEREMOVECOMPLETE:
			switch (lpdb -> dbch_devicetype) {
			case DBT_DEVTYP_DEVICEINTERFACE:
				Log(TEXT("WM_DEVICECHANGE::DBT_DEVICEREMOVECOMPLETE::DBT_DEVTYP_DEVICEINTERFACE"));
				ListDevices();
				break;
			default:
				Log(TEXT("WM_DEVICECHANGE::DBT_DEVICEREMOVECOMPLETE"));
				return FALSE;
			}
			break;

		default:
			Log(TEXT("WM_DEVICECHANGE"));
			return FALSE;
		}
		break;

	case WM_INITDIALOG:
		OnInitDialog(hwnd);
		break;

	case WM_CLOSE:
		if (g_bProcessing) {
			Log(TEXT("WM_CLOSE: enabling global close flag - application will close when processing is complete."));
			g_bClose = TRUE;
			return FALSE;
		}
		if (g_hThread) CloseHandle(g_hThread);
		DestroyList();
		UnregisterDeviceNotification(g_hDevNotify);
		AnimateWindow(hwnd, 200, AW_CENTER | AW_HIDE);
		EndDialog(hwnd, FALSE);
		break;

	default:
		return FALSE;
	}

	return TRUE;
}


BOOL CALLBACK AboutDlgProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	return FALSE;
}


BOOL CALLBACK BmpWindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	HDC hdc;
	PAINTSTRUCT ps;
	switch (uMsg)
	{
	case WM_PAINT:
		hdc = BeginPaint(hwnd, &ps);
		OnPaint(hdc);
		EndPaint(hwnd, &ps);
		break;
	case WM_SETCURSOR:
		SetCursor(LoadCursor(NULL, IDC_CROSS));
		break;
	default:
		return DefWindowProc(hwnd, uMsg, wParam, lParam);
	}
	return TRUE;
}


void OnInitDialog(HWND hwnd)
{
	g_hDlg = hwnd;
	g_hBmp = GetDlgItem(hwnd, IDC_BITMAP);
	g_DevList = NULL;
	g_DevListLength = 0;
	g_hThread = NULL;
	g_bProcessing = FALSE;
	g_bClose = FALSE;

	HICON hIcon;
	hIcon = (HICON)LoadImage(g_hInstance, MAKEINTRESOURCE(IDI_MAINICON), IMAGE_ICON, GetSystemMetrics(SM_CXICON), GetSystemMetrics(SM_CYICON), LR_SHARED);
	if(hIcon) {
		SendMessage(hwnd, WM_SETICON, ICON_BIG, (LPARAM)hIcon);
	}
	hIcon = (HICON)LoadImage(g_hInstance, MAKEINTRESOURCE(IDI_MAINICON), IMAGE_ICON, GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), LR_SHARED);
	if(hIcon) {
		SendMessage(hwnd, WM_SETICON, ICON_SMALL, (LPARAM)hIcon);
	}
	ListScreens();
	SetWindowLong(g_hBmp, GWL_WNDPROC, (LONG)BmpWindowProc);
	SendDlgItemMessage(hwnd, IDC_PROGRESS, PBM_SETPOS, 0, 0);

	int nArgs;
	LPTSTR *szArglist;
	szArglist = CommandLineToArgvW(GetCommandLineW(), &nArgs);
	if (nArgs > 1) {
		g_bProcessing = TRUE;
		g_lpDevice = (LPTSTR)LocalAlloc(NULL, sizeof(TCHAR)*(1+lstrlen(szArglist[1])));
		lstrcpy(g_lpDevice, szArglist[1]);
		if (nArgs > 2) {
			lstrcpy(g_lpBaseName, szArglist[2]);
		}
		SendDlgItemMessage(hwnd, IDC_LSTDEVICE, CB_ADDSTRING, 0, (LPARAM)g_lpDevice);
		SendDlgItemMessage(hwnd, IDC_LSTDEVICE, CB_SETCURSEL, 0, 0);
		g_bProcessing = TRUE;
		g_hThread = CreateThread(NULL, 0, ThreadProc, 0, 0, NULL);
		PostMessage(hwnd, WM_CLOSE, NULL, NULL);
	} else {
		DoRegisterDeviceInterface(GUID_DEVINTERFACE_MODEM, &g_hDevNotify);
		UDiskToMode(AT_Mode_Modem);
		ListDevices();
	}
	LocalFree(szArglist);
}


BOOL OnCommand(WPARAM wParam)
{
	UINT nDevice, nScreen;

	switch (LOWORD(wParam))
	{
	case IDC_BTNCAPTURE:
		nDevice = (UINT)SendDlgItemMessage(g_hDlg, IDC_LSTDEVICE, CB_GETCURSEL, 0, 0);
		nScreen = (UINT)SendDlgItemMessage(g_hDlg, IDC_LSTSCREEN, CB_GETCURSEL, 0, 0);
		if (CB_ERR == nDevice || CB_ERR == nScreen) {
			MessageBox(g_hDlg, TEXT("No device selected"), NULL, MB_ICONINFORMATION | MB_OK);
			return FALSE;
		}
		g_lpDevice = g_DevList[nDevice];
		g_nScreen = nScreen;
		if (g_hThread) CloseHandle(g_hThread);
		g_hThread = CreateThread(NULL, 0, ThreadProc, 0, 0, NULL);
		break;

	case IDC_BTNSAVE:
#ifdef UNICODE
		OPENFILENAMEW ofn;
#else
		OPENFILENAME  ofn;
#endif
		ZeroMemory(&ofn, sizeof(ofn));
		ofn.lStructSize       = sizeof(OPENFILENAME);
		ofn.hwndOwner         = g_hDlg;
		ofn.lpstrFilter       = TEXT("PNG\0*.png\0");
		ofn.nFilterIndex      = 0;
		ofn.lpstrFile         = g_lpBaseName;
		ofn.nMaxFile          = sizeof(g_lpBaseName);
		ofn.lpstrFileTitle    = NULL;
		ofn.nMaxFileTitle     = 0;
		ofn.lpstrInitialDir   = NULL;
		ofn.Flags             = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST | OFN_DONTADDTORECENT;
		if (!GetSaveFileName(&ofn)) {
			return FALSE;
		}
		break;

	default:
		return FALSE;
	}

	return TRUE;
}


void OnPaint(HDC hdc)
{
	Gdiplus::Graphics graphics(hdc);
	graphics.DrawImage(GetBitmap(), 0, 0);
}


BOOL DoRegisterDeviceInterface(GUID InterfaceClassGuid, HDEVNOTIFY *hDevNotify)
{
    DEV_BROADCAST_DEVICEINTERFACE NotificationFilter;

    ZeroMemory(&NotificationFilter, sizeof(NotificationFilter));
    NotificationFilter.dbcc_size = sizeof(DEV_BROADCAST_DEVICEINTERFACE);
    NotificationFilter.dbcc_devicetype = DBT_DEVTYP_DEVICEINTERFACE;
    NotificationFilter.dbcc_classguid = InterfaceClassGuid;

    *hDevNotify = RegisterDeviceNotification(g_hDlg,
        &NotificationFilter,
        DEVICE_NOTIFY_WINDOW_HANDLE
    );

    if(!*hDevNotify) {
        MessageBox(g_hDlg, TEXT("RegisterDeviceNotification failed"), TEXT("Error"), MB_ICONERROR | MB_OK);
        return FALSE;
    }

    return TRUE;
}


void ListDevices()
{
	Log(TEXT("Device list: {"));

	SendDlgItemMessage(g_hDlg, IDC_LSTDEVICE, CB_RESETCONTENT, 0, 0);

	DestroyList();

	DWORD i;

	SP_DEVINFO_DATA DeviceInfoData;
	DeviceInfoData.cbSize = sizeof(SP_DEVINFO_DATA);

	SP_DEVICE_INTERFACE_DATA DeviceInterfaceData;
	DeviceInterfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);

	PSP_DEVICE_INTERFACE_DETAIL_DATA pDeviceInterfaceDetailData;
	DWORD DeviceInterfaceDetailDataSize;

	// get all present modems
	HDEVINFO hDevInfo = SetupDiGetClassDevs(
		&GUID_DEVINTERFACE_MODEM,
		NULL,
		NULL,
		DIGCF_DEVICEINTERFACE | DIGCF_PRESENT
	);

	for (i = 0; SetupDiEnumDeviceInfo(hDevInfo, i, &DeviceInfoData); i++);

	g_DevListLength = i;
	g_DevList = (LPTSTR *)LocalAlloc(NULL, g_DevListLength*sizeof(LPTSTR));

	for (i = 0; i < g_DevListLength; i++) {

		LPTSTR lpDevName = NULL;
		DWORD cDevName = 0;

		SetupDiEnumDeviceInfo(hDevInfo, i, &DeviceInfoData);

		//
		// Call function with null to begin with,
		// then use the returned buffer size
		// to Alloc the buffer. Keep calling until
		// success or an unknown failure.
		//
		while (!SetupDiGetDeviceRegistryProperty(
			hDevInfo,
			&DeviceInfoData,
			SPDRP_FRIENDLYNAME,
			NULL,
			(PBYTE)lpDevName,
			cDevName,
			&cDevName))
		{
			if (GetLastError() == ERROR_INSUFFICIENT_BUFFER) {
				// Change the buffer size.
				if (lpDevName) LocalFree(lpDevName);
				lpDevName = (LPTSTR)LocalAlloc(NULL, cDevName);
			} else {
				// Insert error handling here.
				break;
			}
		}

		LPTSTR lpPortName = NULL;
		DWORD cPortName = -1;
		HKEY hKey = SetupDiOpenDevRegKey(hDevInfo, &DeviceInfoData, DICS_FLAG_GLOBAL, 0, DIREG_DEV, KEY_QUERY_VALUE);
		if (hKey != INVALID_HANDLE_VALUE) {
			LONG result;
			LPTSTR regValueNames[] = { TEXT("PortName"), TEXT("AttachedTo") };
			const int nregValueNames = 2;
			int i = 0;
			while (i < nregValueNames) {
				LONG result;
				result = RegQueryValueEx(hKey, regValueNames[i], NULL, NULL, (LPBYTE)lpPortName, &cPortName);
				if (result == ERROR_SUCCESS) {
					if (lpPortName) LocalFree(lpPortName);
					lpPortName = (LPTSTR)LocalAlloc(NULL, cPortName);
					RegQueryValueEx(hKey, regValueNames[i], NULL, NULL, (LPBYTE)lpPortName, &cPortName);
					break;
				} else {
					i++;
				}
			}
			RegCloseKey(hKey);			
		}
		LPTSTR lpFullDevName;
		if (lpPortName) {
			lpFullDevName = (LPTSTR)LocalAlloc(NULL, cDevName + cPortName + sizeof(TCHAR)*2);
			wsprintf(lpFullDevName, TEXT("%s (%s)"), lpDevName, lpPortName);
			LocalFree(lpDevName);
			LocalFree(lpPortName);
		} else {
			lpFullDevName = lpDevName;
		}

		SendDlgItemMessage(g_hDlg, IDC_LSTDEVICE, CB_ADDSTRING, 0, (LPARAM)lpFullDevName);
		Log(lpFullDevName);
		if (lpFullDevName) LocalFree(lpFullDevName);

		SetupDiEnumDeviceInterfaces(
			hDevInfo,
			&DeviceInfoData,
			&GUID_DEVINTERFACE_MODEM,
			0,
			&DeviceInterfaceData
		);
		pDeviceInterfaceDetailData = NULL;
		DeviceInterfaceDetailDataSize = 0;

		while (!SetupDiGetDeviceInterfaceDetail(
				hDevInfo,
				&DeviceInterfaceData,
				NULL,
				DeviceInterfaceDetailDataSize,
				&DeviceInterfaceDetailDataSize,
				NULL))
		{
			if (GetLastError() ==
				ERROR_INSUFFICIENT_BUFFER)
			{
				// Change the buffer size.
				if (pDeviceInterfaceDetailData) LocalFree(pDeviceInterfaceDetailData);
				pDeviceInterfaceDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA)
					LocalAlloc(NULL, DeviceInterfaceDetailDataSize);
				pDeviceInterfaceDetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);
			}
			else
			{
				// Insert error handling here.
				break;
			}
		}

		// obtain device interface details
		SetupDiGetDeviceInterfaceDetail(
			hDevInfo,
			&DeviceInterfaceData,
			pDeviceInterfaceDetailData,
			DeviceInterfaceDetailDataSize,
			NULL,
			NULL
		);

		// get device path
		LPTSTR lpDevPath = pDeviceInterfaceDetailData->DevicePath;

		g_DevList[i] = (LPTSTR)LocalAlloc(NULL, sizeof(TCHAR)*(1+lstrlen(lpDevPath)));
		lstrcpy(g_DevList[i], lpDevPath);
		Log(lpDevPath);

		if (pDeviceInterfaceDetailData) LocalFree(pDeviceInterfaceDetailData);
	}

	SetupDiDestroyDeviceInfoList(hDevInfo);

	UINT nDevices = (UINT)SendDlgItemMessage(g_hDlg, IDC_LSTDEVICE, CB_GETCOUNT, 0, 0);

	if (nDevices && nDevices != CB_ERR)
		SendDlgItemMessage(g_hDlg, IDC_LSTDEVICE, CB_SETCURSEL, (WPARAM)nDevices-1, 0);

	Log(TEXT("}"));
}


void ListScreens()
{
	SendDlgItemMessage(g_hDlg, IDC_LSTSCREEN, CB_ADDSTRING, 0, (LPARAM)TEXT("Primary"));
	SendDlgItemMessage(g_hDlg, IDC_LSTSCREEN, CB_ADDSTRING, 0, (LPARAM)TEXT("Secondary"));
	SendDlgItemMessage(g_hDlg, IDC_LSTSCREEN, CB_SETCURSEL, 0, 0);
}


void DestroyList()
{
	for (UINT i = 0; i < g_DevListLength; i++) {
		if (g_DevList[i]) LocalFree(g_DevList[i]);
	}
	if (g_DevList) LocalFree(g_DevList);
	g_DevList = NULL;
}


DWORD WINAPI ThreadProc(LPVOID lParam)
{
	Log(TEXT("Starting capture"));
	g_bProcessing = TRUE;

	EnableWindow(GetDlgItem(g_hDlg, IDC_BTNCAPTURE), FALSE);
	EnableWindow(GetDlgItem(g_hDlg, IDC_BTNSAVE), FALSE);
	EnableWindow(GetDlgItem(g_hDlg, IDC_LSTDEVICE), FALSE);
	EnableWindow(GetDlgItem(g_hDlg, IDC_LSTSCREEN), FALSE);

	HWND hProgress = GetDlgItem(g_hDlg, IDC_PROGRESS);
	DWORD style = GetWindowLong(hProgress, GWL_STYLE);
	SetWindowLong(hProgress, GWL_STYLE, style | PBS_MARQUEE);
	SendMessage(hProgress, PBM_SETMARQUEE, TRUE, 100);

	if (!serial_open(g_lpDevice)) {

		MessageBox(g_hDlg, TEXT("Error opening device"), TEXT("Error"), MB_ICONERROR | MB_OK);

	} else {
		Log(g_lpDevice);

		if (!get_modem_reply("+MODE=9")) {
			MessageBox(g_hDlg, TEXT("Error entering test mode. Make sure correct device is selected."), TEXT("Error"), MB_ICONERROR | MB_OK);
		} else {

			char AT_Command2[50];
			int x, y, w, h;
			x = y = 0;
			w = h = 999;
			//w = 176;
			//h = 220;
			wsprintfA(AT_Command2, "+MTEST=img,%d,%d,%d,%d,%d", x, y, w, h, g_nScreen);
			if (!send_at_command(AT_Command2)) {
				MessageBox(g_hDlg, TEXT("Error sending command."), TEXT("Error"), MB_ICONERROR | MB_OK);
			}
			if (GetPixels(g_hDlg)) {
				if (!get_modem_reply(NULL)) {
					MessageBox(g_hDlg, TEXT("Error capturing screen."), TEXT("Error"), MB_ICONERROR | MB_OK);
				} else if (!get_modem_reply("+MODE=0")) {
					MessageBox(g_hDlg, TEXT("Error exiting test mode."), TEXT("Error"), MB_ICONERROR | MB_OK);
				} else {
					SaveImage(g_lpBaseName);
				}
			}
		}

		serial_close();
	}

	EnableWindow(GetDlgItem(g_hDlg, IDC_BTNCAPTURE), TRUE);
	EnableWindow(GetDlgItem(g_hDlg, IDC_BTNSAVE), TRUE);
	EnableWindow(GetDlgItem(g_hDlg, IDC_LSTDEVICE), TRUE);
	EnableWindow(GetDlgItem(g_hDlg, IDC_LSTSCREEN), TRUE);

	SendMessage(hProgress, PBM_SETMARQUEE, FALSE, 0);
	SetWindowLong(hProgress, GWL_STYLE, style);

	g_bProcessing = FALSE;

	Log(TEXT("Capture complete"));
	if (g_bClose) {
		Log(TEXT("Global close flag detected - sending WM_CLOSE"));
		SendMessage(g_hDlg, WM_CLOSE, NULL, NULL);
	}

	return 0;
}
