#include "initguid.h"
#include "windows.h"
#include "ntddscsi.h"
#include "setupapi.h"
#include "shlwapi.h"
#include "udisk.h"
#include "utilfunc.h"


BOOL UDiskDetect(HANDLE *phDevice, BOOL bOpen)
{
	Log(TEXT("UDiskDetect(): Detecting UDisk"));
	BOOL bFound = FALSE;

	// get all present mass storage devices
	HDEVINFO hDeviceInfo = SetupDiGetClassDevs(
		&GUID_DEVINTERFACE_DISK,
		NULL,
		NULL,
		DIGCF_DEVICEINTERFACE | DIGCF_PRESENT
	);

	SP_DEVICE_INTERFACE_DATA DeviceInterfaceData;
	DeviceInterfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);

	PSP_DEVICE_INTERFACE_DETAIL_DATA pDeviceInterfaceDetailData;
	DWORD DeviceInterfaceDetailDataSize;

	// enumerate disk device interfaces
	UINT i = 0;
	while (!bFound &&
		SetupDiEnumDeviceInterfaces(
			hDeviceInfo,
			NULL,
			&GUID_DEVINTERFACE_DISK,
			i++,
			&DeviceInterfaceData)) {

		// get required buffer size
		SetupDiGetDeviceInterfaceDetail(
			hDeviceInfo,
			&DeviceInterfaceData,
			NULL,
			0,
			&DeviceInterfaceDetailDataSize,
			NULL
		);

		// allocate buffer
		pDeviceInterfaceDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA)
			LocalAlloc(NULL, DeviceInterfaceDetailDataSize);
		pDeviceInterfaceDetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

		// obtain device interface details
		SetupDiGetDeviceInterfaceDetail(
			hDeviceInfo,
			&DeviceInterfaceData,
			pDeviceInterfaceDetailData,
			DeviceInterfaceDetailDataSize,
			NULL,
			NULL
		);

		// get device path
		LPTSTR lpDevicePath = pDeviceInterfaceDetailData->DevicePath;

		if (StrStrI(lpDevicePath, TEXT("motorola"))) {
			bFound = TRUE;
			Log(lpDevicePath);

			if (bOpen) {
				// open device
				Log(TEXT("UDiskDetect(): opening device"));
				*phDevice = CreateFile(
					lpDevicePath,
					GENERIC_READ | GENERIC_WRITE,
					FILE_SHARE_READ | FILE_SHARE_WRITE,
					NULL,
					OPEN_EXISTING,
					FILE_FLAG_NO_BUFFERING,
					NULL
				);
				if (*phDevice == INVALID_HANDLE_VALUE) {
					Log(TEXT("UDiskDetect(): error opening device"));
					bFound = FALSE;
				}
			}
		}
		LocalFree(pDeviceInterfaceDetailData);
	}
	SetupDiDestroyDeviceInfoList(hDeviceInfo);

	Log(TEXT("UDiskDetect(): detection complete"));
	return bFound;
}



BOOL UDiskToMode(MMODE mode)
{
	HANDLE hDevice;

	if (!UDiskDetect(&hDevice, TRUE)) {
		Log(TEXT("UDiskToMode(): UDisk not found"));
		return FALSE;
	}

	SCSI_PASS_THROUGH_DIRECT srb;

	srb.Length = sizeof(SCSI_PASS_THROUGH_DIRECT);
	srb.PathId = 0;
	srb.TargetId = 0;
	srb.Lun = 0;
	srb.CdbLength = 12;
	srb.SenseInfoLength = 0;
	srb.DataIn = SCSI_IOCTL_DATA_IN;
	srb.DataTransferLength = 0;
	srb.TimeOutValue = 25;
	srb.DataBuffer = 0;
	srb.SenseInfoOffset = 0;

	srb.Cdb[0] = 0xd6;
	srb.Cdb[10] = mode;

	DWORD nBytesReturned;

	BOOL bResult = DeviceIoControl(
		hDevice,
		IOCTL_SCSI_PASS_THROUGH_DIRECT,
		&srb,
		sizeof(SCSI_PASS_THROUGH_DIRECT),
		NULL,
		0,
		&nBytesReturned,
		NULL
	);

	CloseHandle(hDevice);
	if (bResult) {
		Log(TEXT("UDiskToMode(): success"));
	} else {
		Log(TEXT("UDiskToMode(): error"));
	}
	return bResult;
}
